#include <fits.h>
#include "wfc3psfdata.h"

ftype fits,pam;
double RN,EXP,EXP0,EPOCH,FIXED_ET=-1.;
float DMIN,DMAX;
int MASKCR=1,NCOMBINE=1,FIXED_NC=-1;

int WFC3type(ftype *f) {
   int i;
   //undrizzled UVIS
   if (fits.Next==6) {
      for (i=0;i<6 && f->ext[i].X==4096 && f->ext[i].Y==2051 && f->ext[i].Z==1;i++);
      if (i==6) return 1;
   }
   if (fits.Next==3) {
      for (i=1;i<3 && f->ext[i].X==f->ext[0].X && f->ext[i].Y==f->ext[0].Y && (i==200 || f->ext[i].Z==f->ext[0].Z);i++);
      if (i==3 && f->ext[0].Z==1) {
	 //drizzled UVIS
	 if (f->ext[0].X>=4096 && f->ext[0].Y>=4096) return 2;
	 //drizzled IR
	 if (f->ext[0].X>=900 && f->ext[0].Y>=900) return 4;
      }
   }
   if (fits.Next==5) {
      //undrizzled IR
      for (i=0;i<5 && f->ext[i].X==1014 && f->ext[i].Y==1014 && f->ext[i].Z==1;i++);
      if (i==5) return 3;
   }
   return 0;
}

void WFC3getcards(int ext,int MODE,int drz) {
   int x,y;
   char str[81];

   if (strcmp(getcardval(&(fits.img),"FILETYPE",1),"SCI") || strcmp(getcardval(&(fits.img),"TELESCOP",1),"HST") || strcmp(getcardval(&(fits.img),"INSTRUME",1),"WFC3")) {printf("**Format error (filetype,telescop,instrume)\n"); exit(-1);}

   // IR image
   if (MODE==0 || MODE==-2) {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.25*(atof(getcardval(&(fits.img),"READNSEA",1))+atof(getcardval(&(fits.img),"READNSEB",1))+atof(getcardval(&(fits.img),"READNSEC",1))+atof(getcardval(&(fits.img),"READNSED",1)));
   }

   // WFC, FLT or CRJ
   else if (MODE==1) {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.5*(atof(getcardval(&(fits.img),"READNSEC",1))+atof(getcardval(&(fits.img),"READNSED",1)));
   }
   else if (MODE==2) {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.5*(atof(getcardval(&(fits.img),"READNSEA",1))+atof(getcardval(&(fits.img),"READNSEB",1)));
   }

   // WFC DRZ
   else {
      if (strcmp(getcardval(&(fits.img),"CCDAMP",1),"ABCD")) {printf("**Format error (ccdamp)\n"); exit(-1);}
      RN=0.25*(atof(getcardval(&(fits.img),"READNSEA",1))+atof(getcardval(&(fits.img),"READNSEB",1))+atof(getcardval(&(fits.img),"READNSEC",1))+atof(getcardval(&(fits.img),"READNSED",1)));
   }
   if (FIXED_ET>0) EXP=FIXED_ET;
   else EXP=atof(getcardval(&(fits.img),"EXPTIME",1));
   if (drz) {
      DMIN=DMAX=fits.ext[ext].data[0][0][0];
      for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) {
	 if (fits.ext[ext].data[0][y][x]<DMIN) DMIN=fits.ext[ext].data[0][y][x];
	 else if (fits.ext[ext].data[0][y][x]>DMAX) DMAX=fits.ext[ext].data[0][y][x];
      }
   }
   else {
      DMIN=(float)atof(getcardval(fits.ext+ext,"GOODMIN",1))-1;
      DMAX=(float)atof(getcardval(fits.ext+ext,"GOODMAX",1))+1;
   }
   EPOCH=0.5*(atof(getcardval(&(fits.img),"EXPSTART",1))+atof(getcardval(&(fits.img),"EXPEND",1)));
   if (FIXED_NC>0) {
      NCOMBINE=FIXED_NC;
      printf("Setting number of images to %d\n",NCOMBINE);
   }
   else {
      strcpy(str,getcardval(&(fits.img),"NCOMBINE",0));
      if (str[0]) NCOMBINE=atoi(getcardval(&(fits.img),"NCOMBINE",0));
      if (!str[0]) printf("NCOMBINE keyword not found; assuming data came from single image\n");
      else if (NCOMBINE<1) {
	 printf("Illegal NCOMBINE setting; assuming data came from single image\n");
	 NCOMBINE=1;
      }
      else printf("NCOMBINE = %d\n",NCOMBINE);
   }
   EXP0=EXP/NCOMBINE;
   return;
}

// ignoring type 64 (warm pixel) on assumption it is fixed OK
// ignoring type 16384 (ghost/crosstalk)
void WFC3_DQ_mask(ftype*f,int ext,int drz,int ir) {
   int x,y,dq,i;
   for (y=0;y<f->ext[ext].Y;y++) for (x=0;x<f->ext[ext].X;x++) if (!drz) {
      dq=(int)(f->ext[ext+2].data[0][y][x]+0.5);
      if (ir) {
	 // 1024 unused at present
	 if (dq&256) f->ext[ext].data[0][y][x]=DMAX+1; // 256
	 else if (dq&6847) f->ext[ext].data[0][y][x]=DMIN-1; // 4096+2048+512+128+63
	 else if ((dq&8192) && MASKCR) f->ext[ext].data[0][y][x]=DMIN-1;
      }
      else {
	 if (dq&2304) f->ext[ext].data[0][y][x]=DMAX+1; // 2048+256
	 else if (dq&5823) f->ext[ext].data[0][y][x]=DMIN-1; // 4096+1024+512+128+63
	 else if ((dq&8192) && MASKCR) f->ext[ext].data[0][y][x]=DMIN-1; // 8192
      }
   }
   else {
      dq=0;
      for (i=0;i<f->ext[ext].Z && !dq;i++) if ((int)(f->ext[ext+2].data[i][y][x]+0.5)!=0) dq=1;
      if (dq==0) f->ext[ext].data[0][y][x]=DMIN-1;
   }
   return;
}

void WFC3_PAM_mult(int ext,char*pamfn,double mult) {
   int x,y;
   float DMIN1,DMAX1;
   char fn[321];

   if (DMAX>0.) DMAX1=1.5*DMAX;
   else DMAX1=0.;
   if (DMIN<0.) DMIN1=1.5*DMIN;
   else DMIN1=0.;
   sprintf(fn,"%s/wfc3/data/%s",BASEDIR,pamfn);
   readfits(fn,&pam,0);
   if (pam.Next!=1 || pam.ext[0].X!=fits.ext[ext].X || pam.ext[0].Y!=fits.ext[ext].Y || pam.ext[0].Z!=fits.ext[ext].Z) {
      printf("Error in PAM file %s\n",pamfn);
      exit(-1);
      return;
   }
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) if (fits.ext[ext].data[0][y][x]>DMIN && fits.ext[ext].data[0][y][x]<DMAX) {
      fits.ext[ext].data[0][y][x]*=pam.ext[0].data[0][y][x]/mult;
      fits.ext[ext+1].data[0][y][x]*=pam.ext[0].data[0][y][x]/mult;
   }
   else if (fits.ext[ext].data[0][y][x]>=DMAX) fits.ext[ext].data[0][y][x]=DMAX1+1;
   else fits.ext[ext].data[0][y][x]=DMIN1-1;
   freefits(&pam);
   DMIN=DMIN1;
   DMAX=DMAX1;
   return;
}

void WFC3_exptime_mult(int ext,double mult) {
   int x,y;

   DMAX*=EXP/mult;
   DMIN*=EXP/mult;
   for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) {
      fits.ext[ext].data[0][y][x]*=EXP/mult;
      fits.ext[ext+1].data[0][y][x]*=EXP/mult;
   }
   return;
}

/*
  auto-solve for noise parameters:
  y = noise^2 = rdnoise^2/gain^2 + data/gain = a + b*data
  chi^2 = SUM[(a+b*x-y)^2/sig^2] = SUM[(a/y+b*x/y-1)^2]
  dchi^2/da = 0 = SUM[(a/y+b*x/y-1)/y]
  dchi^2/db = 0 = SUM[(a/y+b*x/y-1)*x/y]
  I=SUM(1/yy); X=SUM(x/yy); Y=SUM(1/y); XX=SUM(xx/yy); XY=SUM(x/y);
  0 = a*I+b*X-Y   = a*I*X + b*X*X - X*Y   = a*XX*I + b*X*XX - XX*Y
  0 = a*X+b*XX-XY = a*I*X + b*I*XX - I*XY = a*X*X + b*X*XX - X*XY
  b = (I*XY-X*Y)/(I*XX-X*X) = 1/gain
  a = (XX*Y-X*XY)/(I*XX-X*X) = rdnoise^2/gain^2
  rdnoise = gain*sqrt((XX*Y-X*XY)/(I*XX-X*X))

  assuming gain=1 (should be the case):
  rdnoise^2 is simply the average of noise^2-data
  0 = a*I+X-Y  -->  a=(Y-X)/I
*/
void WFC3setcards(int ext,int ir) {
   int i;
   /*
   if (ir)
   {
      int x,y;
      chiptype data,noise;
      double I=0,X=0,Y=0,XX=0,XY=0,tGN,tRN;

      data=fits.ext[ext].data[0];
      noise=fits.ext[ext+1].data[0];
      for (y=0;y<fits.ext[ext].Y;y++) for (x=0;x<fits.ext[ext].X;x++) if (data[y][x]>DMIN && data[y][x]<DMAX) {
	 int dq=(int)(fits.ext[ext+2].data[0][y][x]+0.5);
	 if ((dq&256)==0 && (dq&8192)==0) {
	    double isig2;
	    isig2=1./(noise[y][x]*noise[y][x]*noise[y][x]*noise[y][x]);
	    printf("%f %f ^^^^\n",noise[y][x],data[y][x]);
	    I+=isig2;
	    Y+=noise[y][x]*noise[y][x]*isig2;
	    if (data[y][x]>0.) {
	       X+=data[y][x]*isig2;
	       XX+=data[y][x]*data[y][x]*isig2;
	       XY+=data[y][x]*noise[y][x]*noise[y][x]*isig2;
	    }
	 }
      }
      tGN=(XX*I-X*X)/(XY*I-X*Y);
      tRN=sqrt((XX*Y-X*XY)/(XX*I-X*X))*tGN;
      printf("%f %f\n",tGN,tRN);
   }
   */
   insertcards(fits.ext+ext,1.,RN*sqrt(NCOMBINE),EXP,DMIN,DMAX,EPOCH,0.0,EXP0);
   freeim(fits.ext+ext+1);
   freeim(fits.ext+ext+2);
   if (ir && fits.Next==5) {
      freeim(fits.ext+ext+3);
      freeim(fits.ext+ext+4);
      fits.Next-=4;
      for (i=ext+1;i<fits.Next;i++) memcpy(fits.ext+i,fits.ext+i+4,sizeof(imtype));
   }
   else {
      fits.Next-=2;
      for (i=ext+1;i<fits.Next;i++) memcpy(fits.ext+i,fits.ext+i+2,sizeof(imtype));
   }
   return;
}

int main(int argc,char**argv) {
   int i,tp;

   if (argc<2) {
      printf("Usage: %s <<-flags>> <fits files>\n",*argv);
      printf(" -keepcr      leaves fixed cosmic ray pixels in image\n");
      printf(" -exptime=#   overrides exposure time; needed for _drz images\n");
      printf(" -ncombine=#  overrides NCOMBINE keyword\n");
      return 1;
   }
   for (i=1;i<argc;i++) if (!strcasecmp(argv[i],"-KEEPCR")) MASKCR=0;
   else if (!strncasecmp(argv[i],"-EXPTIME=",9)) FIXED_ET=atof(argv[i]+9);
   else if (!strncasecmp(argv[i],"-NCOMBINE=",10)) FIXED_NC=atof(argv[i]+10);
   else {
      readfits(argv[i],&fits,1);
      tp=WFC3type(&fits);
      if (fits.Next>0 && strcmp(getcardval(fits.ext,"DOL_WFC3",0),"")) printf("%s already run through wfc3mask\n",argv[i]);
      else if (tp<1 || tp>4) printf("%s is not a WFC3 fits file\n",argv[i]);
      else {
	 if (tp==1) { // FLT or CRJ UVIS image
	    WFC3getcards(0,2,0);
	    WFC3_DQ_mask(&fits,0,0,0);
	    WFC3_PAM_mult(0,"UVIS2wfc3_map.fits",wfc3_ctmult[2]);
	    WFC3setcards(0,0);
	    WFC3getcards(1,1,0);
	    WFC3_DQ_mask(&fits,1,0,0);
	    WFC3_PAM_mult(1,"UVIS1wfc3_map.fits",wfc3_ctmult[1]);
	    WFC3setcards(1,0);
	    addcard(fits.ext,"DOL_WFC3=                    2 / DOLPHOT WFC3 tag                               ");
	    addcard(fits.ext+1,"DOL_WFC3=                    1 / DOLPHOT WFC3 tag                               ");
	    fits.Next=2;
	 }
	 else if (tp==2) { // DRZ UVIS image
	    WFC3getcards(0,-1,1);
	    WFC3_DQ_mask(&fits,0,1,0);
	    WFC3_exptime_mult(0,wfc3_ctmult[1]);
	    WFC3setcards(0,0);
	    fits.Next=1;
	    addcard(fits.ext,"DOL_WFC3=                   -2 / DOLPHOT WFC3 tag                               ");
	 }
	 else if (tp==3) { // FLT or CRJ IR image
	    WFC3getcards(0,0,0);
	    WFC3_DQ_mask(&fits,0,0,1);
	    WFC3_PAM_mult(0,"ir_wfc3_map.fits",wfc3_ctmult[0]);
	    WFC3_exptime_mult(0,wfc3_ctmult[0]);
	    WFC3setcards(0,1);
	    fits.Next=1;
	    addcard(fits.ext,"DOL_WFC3=                    0 / DOLPHOT WFC3 tag                               ");
	 }
	 else if (tp==4) { // DRZ IR image
	    WFC3getcards(0,-2,1);
	    WFC3_DQ_mask(&fits,0,1,1);
	    WFC3_exptime_mult(0,wfc3_ctmult[0]);
	    WFC3setcards(0,1);
	    fits.Next=1;
	    addcard(fits.ext,"DOL_WFC3=                   -1 / DOLPHOT WFC3 tag                               ");
	 }
	 writefits(argv[i],&fits,1);
      }
      freefits(&fits);
   }
   return 0;
}
